/*
Copyright (C) 2006  The PARI group.

This file is part of the PARI package.

PARI/GP is free software; you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation. It is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY WHATSOEVER.

Check the License for details. You should have received a copy of it, along
with the package; see the file 'COPYING'. If not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */

#include "pari.h"
#include "paripriv.h"
#include "anal.h"
#include "tree.h"

void compilenode(int n, int mode);

THREAD char *gp_function_name=NULL;

typedef enum {PPend,PPstd,PPdefault,PPdefaultmulti,PPstar,PPauto} PPproto;

static PPproto
parseproto(char const **q, char *c)
{
  char  const *p=*q;
  long i;
  switch(*p)
  {
  case 0:
  case '\n':
    return PPend;
  case 'D':
    switch(p[1])
    {
    case 0:
      pari_err(talker,"incomplete prototype");
    case 'G':
    case '&':
    case 'V':
    case 'I':
    case 'n':
      *c=p[1];
      *q=p+2;
      return PPdefault;
    default:
      for(i=0;*p && i<2;p++) i+=*p==',';
      if (i<2)
        pari_err(talker,"incomplete prototype");
      *c=p[-2];
      *q=p;
      return PPdefaultmulti;
    }
    break;
  case 'p':
  case 'P':
    *c=*p;
    *q=p+1;
    return PPauto;
  case '&':
    *c='*';
    *q=p+1;
    return PPstd;
  case 'V':
    if (p[1]=='=')
    {
      *c=*p;
      p+=2;
      *q=p+1;
      return PPstar;
    }
  case 's': /*fall through*/
    if (p[1]=='*')
    {
      *c=*p++;
      *q=p+1;
      return PPstar;
    }
  default:/*fall through*/
    *c=*p;
    *q=p+1;
    return PPstd;
  }
}
/*supported types:
 * type: Gsmall, Ggen, Gvar, Gvoid, Gvec
 * mode: Gsmall, Ggen, Gvar, Gvoid
 */
static void
compilecast(int n, int type, int mode)
{
  if (type==mode) return;
  switch (mode)
  {
  case Gsmall:
    if (type==Ggen)        op_push(OCitos,0);
    else if (type==Gvoid)  op_push(OCpushlong,0);
    else pari_err(talker2,"this should be an integer",
             tree[n].start, get_origin());
    break;
  case Ggen:
    if (type==Gsmall)      op_push(OCstoi,0);
    else if (type==Gvoid)  op_push(OCpushlong,(long)gnil);
    else if (type==Gvar )  op_push(OCpolx,0);
    break;
  case Gvoid:
    op_push(OCpop, 1);
    break;
  case Gvar:
    if (type==Ggen)        op_push(OCvarn,0);
    else pari_err(varer1, tree[n].start, get_origin());
     break;
  default:
    pari_err(bugparier,"compilecast, type unknown %ld",mode);
  }
}

static entree *
getentry(int n, long func)
{
  int x=tree[n].x;
  char *name=(char *)value[x].val.str;
  entree *ep = is_entry(name);
  if (!ep)
    ep=fetch_named_var(name);
  else if (!func && ep->valence!=EpVAR && ep->valence!=EpGVAR)
    pari_err(varer1, tree[n].start, get_origin());
  return ep;
}

static GEN 
listtogen(int n, int f)
{
  GEN z;
  int x,i,nb;
  if (n==-1 || tree[n].f==Fnoarg) return cgetg(1,t_VECSMALL);
  for(x=n, i=0; tree[x].f==f ;x=tree[x].x, i++);
  nb=i+1;
  z=cgetg(nb+1, t_VECSMALL);
  for (x=n; i>0; z[i+1]=tree[x].y, x=tree[x].x, i--);
  z[1]=x;
  return z;
}

static entree *
getlvalue(int n)
{ 
  while (tree[n].f==Ffacteurmat)
    n=tree[n].x;
  return getentry(n,0);
}

static void
compilelvalue(int n)
{ 
  if (tree[n].f==Fentry)
    return;
  else
  {
    int x=tree[n].x;
    int y=tree[n].y;
    int yx=tree[y].x;
    int yy=tree[y].y;
    long f=tree[y].f;
    if (tree[x].f==Ffacteurmat && f==Fmatrix && yy==-1 && 
        tree[tree[x].y].f==FmatrixL)
    {
      compilelvalue(tree[x].x);
      compilenode(tree[tree[x].y].x,Gsmall);
      compilenode(yx,Gsmall);
      op_push(OCcompo2ptr,0);
      return;
    }
    compilelvalue(x);
    compilenode(yx,Gsmall);
    if (f==Fmatrix && yy==-1)
      op_push(OCcompo1ptr,0);
    else
    {
      switch(f)
      {
      case Fmatrix:
        compilenode(yy,Gsmall);
        op_push(OCcompo2ptr,0);
        break;
      case FmatrixR:
        op_push(OCcompoCptr,0);
        break;
      case FmatrixL:
        op_push(OCcompoLptr,0);
        break;
      }
    }
  } 
}

static void 
compilefacteurmat(int n, int mode)
{ 
  int x=tree[n].x;
  int y=tree[n].y;
  int yx=tree[y].x;
  int yy=tree[y].y;
  long f=tree[y].f;
  compilenode(x,Ggen);
  compilenode(yx,Gsmall);
  if (f==Fmatrix && yy==-1)
  {
    op_push(OCcompo1,mode);
    return;
  }
  switch(f)
  {
  case Fmatrix:
    compilenode(yy,Gsmall);
    op_push(OCcompo2,mode);
    break;
  case FmatrixR:
    op_push(OCcompoC,0);
    compilecast(n,Gvec,mode);
    return;
  case FmatrixL:
    op_push(OCcompoL,0);
    compilecast(n,Gvec,mode);
    return;
  default:
    pari_err(bugparier,"compilefacteurmat");
  }
}

static GEN
cattovec(int n, int fnum)
{
  int x=n, i=0, nb;
  GEN stack;
  if (tree[n].f==Fnoarg) return cgetg(1,t_VECSMALL);
  while(1)
  {
    int xx=tree[x].x;
    int xy=tree[x].y;
    if (tree[x].f!=Ffunction || xx!=fnum) break;
    x=tree[xy].x;
    i++;
  }
  nb=i+1;
  stack=cgetg(nb+1,t_VECSMALL);
  for(x=n;i>0;i--)
  {
    int y=tree[x].y;
    x=tree[y].x;
    stack[i+1]=tree[y].y;
  }
  stack[1]=x;
  return stack;
}

static entree *
getfunc(int n)
{
  int x=tree[n].x;
  char *name=(char *)value[x].val.str;
  entree *ep; 
  ep = is_entry(name);
  if (!ep)
  {
    ep=fetch_named_var(name);
    ep->valence=EpNEW;
  }
  return do_alias(ep);
}

static void 
compilefunc(int n, int mode)
{
  pari_sp ltop=avma;
  long i,j;
  int x=tree[n].x;
  char *name=(char *)value[x].val.str;
  int ret;
  char const *p,*q;
  char c;
  PPproto mod;
  GEN arg=listtogen(tree[n].y,Flistarg);
  long nbpointers=0;
  long nb=lg(arg)-1;
  entree *ep = getfunc(n);
  if (EpVALENCE(ep)==EpVAR || EpVALENCE(ep)==EpGVAR)
    pari_err(talker2,"not a function in function call",
        tree[n].start, get_origin());
  if (EpVALENCE(ep)==EpUSER|| EpVALENCE(ep)==EpNEW)
  {
    for (j=1;j<=nb;j++)
      if (tree[arg[j]].f!=Fnoarg)
        compilenode(arg[j], Ggen);
      else
        op_push(OCpushlong,0);
    op_push(OCpushlong, nb);
    if (tree[n].f==Fderfunc)
      op_push(OCderivuser, (long) ep);
    else
      op_push(OCcalluser, (long) ep);
    compilecast(n,Ggen,mode);
    avma=ltop;
    return;
  }
  if (!strcmp(name,"local"))
  {
    if (tree[n].f==Fderfunc)
      pari_err(talker2,"can't derive this",tree[n].start,get_origin());
    compilecast(n,Gvoid,mode);
    avma=ltop;
    return;
  }
  if (!strcmp(name,"global"))
  {
    if (tree[n].f==Fderfunc)
      pari_err(talker2,"can't derive this",tree[n].start,get_origin());
    for (i=1;i<lg(arg);i++)
    {
      long a=arg[i];
      long en;
      if (tree[a].f==Faffect)
      {
        compilenode(tree[a].y,Ggen);
        a=tree[a].x;
      }
      else
        op_push(OCpushlong,0);
      if (tree[a].f==Ftag)
        a=tree[a].x;
      en=(long)getentry(a,0);
      op_push(OCglobalvar,en);
    }
    compilecast(n,Gvoid,mode);
    avma=ltop;
    return;
  }
  if (!strcmp(name,"O"))
  {
    int a=tree[n].y;
    if (tree[n].f==Fderfunc)
      pari_err(talker2,"can't derive this",tree[n].start,get_origin());
    if (tree[a].f!=Ffunction || tree[a].x!=OPpow)
    {
      compilenode(a,Ggen);
      op_push(OCpushlong,1);
    }
    else
    {
      int y=tree[a].y;
      compilenode(tree[y].x,Ggen);
      compilenode(tree[y].y,Gsmall);
    }
    op_push(OCcallgen,(long)ep);
    compilecast(n,Ggen,mode);
    avma=ltop;
    return;
  }
  if (x==OPcat)
    pari_err(talker2,"expected character: ',' or ')' instead of",
        tree[arg[1]].end, get_origin());
  p=ep->code;
  if (!ep->value)
    pari_err(talker2,"unknown function",tree[n].end, get_origin());
  if (*p == 'v') { ret = RET_VOID; p++; }
  else if (*p == 'i') { ret = RET_INT;  p++; }
  else if (*p == 'l') { ret = RET_LONG; p++; }
  else ret = RET_GEN;
  if (tree[n].f==Fderfunc && (ret!=RET_GEN || *p!='G'))
    pari_err(talker2,"can't derive this",tree[n].start,get_origin());
  i=0; j=1;
  if (*p)
  {
    q=p; 
    while((mod=parseproto(&p,&c))!=PPend)
    {
      if (j<=nb && tree[arg[j]].f!=Fnoarg 
                && (mod==PPdefault || mod==PPdefaultmulti))
        mod=PPstd;
      switch(mod)
      {
      case PPstd:
        if (j>nb)
          pari_err(talker2,"too few arguments:",
              tree[n].end-1, get_origin());
        if (tree[arg[j]].f==Fnoarg)
          pari_err(talker2,"missing mandatory argument",
              tree[arg[j]].end, get_origin());
        switch(c)
        {
        case 'G':
          compilenode(arg[j++],Ggen);
          break;
        case 'M':
        case 'L':
          compilenode(arg[j++],Gsmall);
          break;
        case 'n':
          compilenode(arg[j++],Gvar);
          break;
        case '&': case '*': 
          {
            int a=arg[j++];
            entree *ep;
            if (c=='&')
            {
              if (tree[a].f!=Frefarg)
                pari_err(talker2,"expected character: '&'",
                    tree[a].start, get_origin());
              a=tree[a].x;
            }
            ep=getlvalue(a);
            op_push(OCnewptr, (long) ep);
            compilelvalue(a);
            op_push(OCpushptr, 0);
            nbpointers++;
            break;
          }
        case 'I':
        case 'E':
          {
            long sop,sdat,off;
            getcodepos(&sop,&sdat,&off);
            compilenode(arg[j++],Ggen);
            op_push(OCpushgen, data_push(getclosure(sop,sdat,off)));
            break;
          }
        case 'V':
        case 'S':
          {
            entree *ep = getentry(arg[j++],c=='S');
            op_push(OCpushlong, (long)ep);
            break;
          }
        case 'r':
          {
            int a=arg[j++];
            if (tree[a].f==Fentry)
            {
              op_push(OCpushgen, 
                  data_push(strtoGENstr(value[tree[a].x].val.str)));
              op_push(OCtostr, 1);
            }
            else
            {
              compilenode(a,Ggen);
              op_push(OCtostr, 1);
            }
            break;
          }
        case 's':
          {
            GEN g = cattovec(arg[j++], OPcat);
            long l, nb = lg(g)-1;
            for(l=1; l<=nb; l++)
              compilenode(g[l], Ggen);
            op_push(OCtostr, nb);
            break;
          }
        default:
          pari_err(talker,"Unknown prototype code `%c' for `%s'",c,name);
        }
        break;
      case PPauto:
        switch(c)
        {
        case 'p':
          op_push(OCprecreal,0);
          break;
        case 'P':
          op_push(OCprecdl,0);
          break;
        }
        break;
      case PPdefault:
        j++;
        switch(c)
        {
        case 'G':
        case '&':
        case 'V':
        case 'r':
        case 'E':
        case 'I':
          op_push(OCpushlong,0);
          break;
        case 'n':
          op_push(OCpushlong,-1);
          break;
        default:
          pari_err(talker,"Unknown default in prototype code `%c' for `%s'",c,name);
        }
        break;
      case PPdefaultmulti:
        j++;
        switch(c)
        {
        case 'L':
        case 'M':
          op_push(OCpushlong,strtol(q+1,NULL,10));
          break;
        case 'r':
        case 's':
          if (q[1]=='"' && q[2]=='"')
            op_push(OCpushlong,(long)"");
          else
            pari_err(impl,"prototype not supported");
          break;
        default:
          pari_err(talker,"Unknown default in prototype code `%c' for `%s'",c,name);
        }
        break;
      case PPstar:
        switch(c)
        {
        case 'V':
          {
            int x=tree[arg[j]].x;
            int y=tree[arg[j]].y;
            entree *ep;
            if (tree[arg[j]].f!=Faffect)
              pari_err(talker2,"expected character: '=' instead of",
                  tree[arg[j]].end, get_origin());
            ep = getentry(x,0);
            op_push(OCpushlong, (long)ep);
            compilenode(y,Ggen);
            i++; j++;
          }
          break;
        case 's':
          {
            long n=nb+1-j;
            long k,l,l1,m;
            GEN g=cgetg(n+1,t_VEC);
            for(l1=0,k=1;k<=n;k++)
            {
              gel(g,k)=cattovec(arg[j+k-1],OPcat);
              l1+=lg(g[k])-1;
            }
            for(m=1,k=1;k<=n;k++)
              for(l=1;l<lg(g[k]);l++)
                compilenode(mael(g,k,l),Ggen);
            j=nb+1;
            op_push(OCvec, l1);
            break;
          }
        default:
          pari_err(talker,"Unknown prototype code `%c*' for `%s'",c,name);
        }
        break;
      default:
        pari_err(bugparier,"PPproto %d in gencallfunc",mod);
      }
      i++;
      q=p;
    }
  }
  if (j<=nb)
    pari_err(talker2,"too many arguments",tree[arg[j]].start,get_origin());
  switch (ret)
  {
  case RET_GEN: 
    if (tree[n].f==Fderfunc)
      op_push(OCderivgen, (long) ep);
    else
      op_push(OCcallgen, (long) ep);
    compilecast(n,Ggen,mode);
    break;
  case RET_INT:
    op_push(OCcallint, (long) ep);
    compilecast(n,Gsmall,mode);
    break;
  case RET_LONG:
    op_push(OCcalllong, (long) ep);
    compilecast(n,Gsmall,mode);
    break;
  case RET_VOID:
    op_push(OCcallvoid, (long) ep);
    compilecast(n,Gvoid,mode);
    break;
  }
  if (nbpointers) op_push(OCendptr,nbpointers);
  avma=ltop;
}

static void
compilevec(int n)
{
  pari_sp ltop=avma;
  int x=tree[n].x;
  long i;
  GEN arg=listtogen(x,Fmatrixelts);
  long l=lg(arg)-1;
  for(i=1;i<=l;i++)
    compilenode(arg[i],Ggen);
  op_push(OCvec,l);
  avma=ltop;
}

static void
compilemat(int n)
{
  pari_sp ltop=avma;
  int x=tree[n].x;
  long i,j;
  GEN line=listtogen(x,Fmatrixlines);
  long nbline = lg(line)-1, nbcol=0;
  GEN lines=cgetg(nbline+1,t_MAT);
  for(i=1;i<=nbline;i++)
  {
    long k;
    lines[i]=(long)listtogen(line[i],Fmatrixelts);
    k=lg(lines[i])-1;
    if (i>1 && k!=nbcol)
      pari_err(talker2,"matrix must be rectangular",
          tree[line[i]].start,get_origin());
    nbcol=k;
  }
  for(i=1;i<=nbcol;i++)
  {
    for(j=1; j<=nbline; j++)
      compilenode(mael(lines,j,i), Ggen);
    op_push(OCcol,nbline);
  }
  op_push(OCmat,nbcol);
  avma=ltop;
}

void
compilenode(int n, int mode)
{
  int x,y;
  if (n<0)
    pari_err(bugparier,"compilenode");
  x=tree[n].x;
  y=tree[n].y;
   
  switch(tree[n].f)
  {
  case Fseq:
    if (tree[x].f!=Fnoarg)
      compilenode(x,Gvoid);
    compilenode(y,mode);
    return;
  case Ffacteurmat:
    compilefacteurmat(n,mode);
    break;
  case Faffect:
    {
      struct node_loc l;
      l.start=tree[n].start;
      l.end=tree[n].end;
      compilefunc(newopcall(OPstore,x,y,&l),mode);
    }
    break;
  case Fconst:
    if (value[x].type!=CSTquote)
    {
      if (mode==Gvoid) return;
      if (mode==Gvar)
        pari_err(varer1,tree[n].start,get_origin());
    }
    if (mode==Gsmall)
      pari_err(talker2,"this should be a small integer",
          tree[n].start,get_origin());
    switch(value[x].type)
    {
    case CSTreal: 
      op_push(OCpushreal, data_push(gcopy(value[x].val.gen)));
      return;
    case CSTgen: 
      op_push(OCpushgen, data_push(gcopy(value[x].val.gen)));
      return;
    case CSTstr:
      op_push(OCpushgen, data_push(strtoGENstr((char *)value[x].val.str)));
      return;
    case CSTquote:
      op_push(OCpushlong, fetch_user_var((char *)value[x].val.str));
      compilecast(n,Gvar, mode);
      return;
    }
    break;
  case Fsmall:
    if (mode==Ggen)
      op_push(OCpushstoi, x);
    else
    {
      op_push(OCpushlong, x);
      compilecast(n,Gsmall,mode);
    }
    return;
  case Fvec:
    compilevec(n);
    compilecast(n,Gvec,mode);
    return;
  case Fmat:
    compilemat(n);
    compilecast(n,Gvec,mode);
    return;
  case Frefarg:
    pari_err(talker,"unexpected &"); 
    break;
  case Fentry:
    {
      char *name=(char *)value[x].val.str;
      entree *ep=is_entry(name);
      if (ep && EpVALENCE(ep)!=EpVAR && EpVALENCE(ep)!=EpGVAR)
        compilefunc(n,mode);
      else
      {
        op_push(OCpushentryval,(long)getentry(n,0));
        compilecast(n,Ggen,mode);
      }
      return;
    }
  case Fderfunc:
  case Ffunction:
    compilefunc(n, mode);
    return;
  case Fdeffunc:
    {
      pari_sp ltop=avma;
      long sop,sdat,off;
      long i;
      GEN arg2=listtogen(tree[x].y,Flistarg);
      entree *ep=getfunc(x);
      int loc=y;
      GEN vars=NULL,lvar;
      long nbvar;
      long arity=lg(arg2)-1;
      if (loc>=0)
        while (tree[loc].f==Fseq) loc=tree[loc].x;
      if (ep->valence!=EpNEW && ep->valence!=EpUSER)
      {
        if (ep->valence==EpVAR || ep->valence==EpGVAR)
          pari_err(talker2,"this is a variable",
              tree[n].start,get_origin());
        else
          pari_err(talker2,"cannot redefine GP functions",
              tree[n].start,get_origin());
      }
      if (loc>=0 && tree[loc].f==Ffunction 
          && strcmp(value[tree[loc].x].val.str,"local")==0)
        vars=listtogen(tree[loc].y,Flistarg);
      nbvar=arity+(vars?lg(vars)-1:0);
      lvar=cgetg(nbvar+1,t_VECSMALL);
      getcodepos(&sop,&sdat,&off);
      for (i=1;i<=arity;i++)
      {
        long a = arg2[lg(arg2)-i];
        long en;
        switch (tree[a].f)
        {
        case Ftag:
          a=tree[a].x;
        case Fentry: /*Fall through*/
          en=(long)getentry(a,0);
          op_push(OCgetarg,en);
          lvar[i]=en;
          break;
        case Faffect:
          { 
            long lop,ldat,loff;
            getcodepos(&lop,&ldat,&loff);
            compilenode(tree[a].y,Ggen);
            a=tree[a].x;
            if (tree[a].f==Ftag)
              a=tree[a].x;
            op_push(OCpushgen, data_push(getclosure(lop,ldat,loff)));
            en=(long)getentry(a,0);
            op_push(OCdefaultarg,en);
            lvar[i]=en;
            break;
          }
        default: 
          pari_err(talker2,"invalid function definition",
              tree[a].start,get_origin());
        }
      }
      if (vars)
      {
        for (i=1;i<lg(vars);i++)
        {
          long a=vars[i];
          long en;
          if (tree[a].f==Faffect)
          {
            compilenode(tree[a].y,Ggen);
            a=tree[a].x;
          }
          else
            op_push(OCpushstoi,0);
          if (tree[a].f==Ftag)
            a=tree[a].x;
          en=(long)getentry(a,0);
          op_push(OCgetarg,en);
          lvar[i+arity]=en;
        }
      }
      if (y>=0 && tree[y].f!=Fnoarg) compilenode(y,Ggen);
      else compilecast(n,Gvoid,Ggen);
      op_push(OCpushgen, data_push(getclosure(sop,sdat,off)));
      op_push(OCpushgen, data_push(lvar));
      op_push(OCpushgen, data_push(
            strntoGENstr(tree[n].start,tree[n].end-tree[n].start)));
      op_push(OCpushlong, arity);
      op_push(OCdeffunc, (long) ep);
      compilecast(n,Gvoid,mode);
      avma=ltop;
      break;
    }
  case Ftag:
    compilenode(x, mode);
    return;
  default:
    pari_err(bugparier,"compilenode");
  }
}

GEN gp_closure(int n)
{
  compilenode(n,Ggen);
  return getclosure(0,0,-1); 
}

struct derivgenwrap_s 
{
  GEN (*f)(ANYARG);
  GEN *x;
};

static GEN
derivgenwrap(GEN x, void* E)
{
  struct derivgenwrap_s *c=(struct derivgenwrap_s *) E;
  GEN z = c->f(x,c->x[1],c->x[2],c->x[3],c->x[4],c->x[5],c->x[6],c->x[7]);
  if (!z) 
    /*This cannot actually happen since no functions returning NULL
     *take a GEN as first parameter*/
    pari_err(talker, "break not allowed here");
  return z;
}

static long
check_array_index(long c, long max)
{
  if (c < 1 || c >= max)
  {
    char s[80];
    sprintf(s,"array index (%ld) out of allowed range ",c);
    if (max == 1) strcat(s, "[none]");
    else if (max == 2) strcat(s, "[1]");
    else sprintf(s,"%s[1-%ld]",s,max-1);
    pari_err(talker,s);
  }
  return c;
}

INLINE long
closure_varn(GEN x)
{
  if (typ(x) != t_POL || lg(x) != 4 ||
      !gcmp0(gel(x,2)) || !gcmp1(gel(x,3))) 
    pari_err(varer1,"variable name expected",NULL,NULL);
  return varn(x);
}

INLINE void
copyvalue(entree *ep, GEN x) {
  new_val_cell(ep, x, typ(x) >= t_VEC ? COPY_VAL: PUSH_VAL);
}
INLINE void
pushvalue(entree *ep, GEN x) { new_val_cell(ep, x, PUSH_VAL); }

INLINE GEN
copyifclone(GEN z)
{
  if (isclone(z)) return gcopy(z);
  else return z;
}

#define killvalue(v) pop_val(get_ep(v))

static long st[1024];
static long sp=0;

static GEN
derivuserwrap(GEN x, void* E)
{
  entree *ep=(entree*)E;
  GEN z;
  long savsp=sp;
  long j;
  gel(st,sp)=x;
  for (j=1;j<ep->arity;j++)
    gel(st,sp+j)=gel(st,sp+j-ep->arity);
  sp+=ep->arity;
  z = closure_eval((GEN) ep->value);
  if (br_status)
  {
    sp = savsp;
    br_status = br_NONE;
    if (br_res) z=gcopy(br_res);
    if (!z) pari_err(talker, "break when derivating");
  }
  if (isclone(z)) z = gcopy(z);
  for(j=1;j<lg(ep->lvars);j++)
    pop_val((entree*)ep->lvars[j]);
  return z;
}


INLINE void
closure_castgen(GEN z, long mode)
{
  switch (mode)
  {
  case Ggen:
    gel(st,sp++)=gcopy(z);
    break;
  case Gsmall:
    st[sp++]=itos(z);
    break;
  case Gvar:
    st[sp++]=closure_varn(z);
    break;
  case Gvoid:
    break;
  default:
    pari_err(bugparier,"closure_castgen, type unknown");
  }
}

INLINE void
closure_castlong(long z, long mode)
{
  switch (mode)
  {
  case Gsmall:
    st[sp++]=z;
    break;
  case Ggen:
    gel(st,sp++)=stoi(z);
    break;
  case Gvar:
    pari_err(varer1,"variable name expected",NULL,NULL);
  default:
    pari_err(bugparier,"closure_castlong, type unknown");
  }
}

GEN
closure_eval(GEN C)
{
  pari_sp ltop=avma;
  gp_pointer ptrs[16];
  long rp=0;
  GEN code=gel(C,1);
  GEN data=gel(C,2);
  long pc,j;
  for(pc=1;pc<lg(code);pc+=2)
  {
    op_code opcode=code[pc];
    long operand=code[pc+1];
    entree *ep;
    if (sp>1000) pari_err(talker,"evaluator stack exhausted");
    switch(opcode)
    {
    case OCpushlong:
        st[sp++]=operand;
        break;
    case OCpushgen:
        gel(st,sp++)=gel(data,operand);
        break;
    case OCpushreal:
        gel(st,sp++)=strtor(GSTR(data[operand]),precreal);
        break;
    case OCpushstoi:
        gel(st,sp++)=stoi(operand);
        break;
    case OCpushentryval:
        ep=(entree*)operand;
        gel(st,sp++)=(GEN)ep->value;
        break;
    case OCnewptr:
        if (rp<16)
        {
          gp_pointer *g = &ptrs[rp++];
          g->ep = (entree*) operand;
          g->x = (GEN) g->ep->value;
          break;
        }
        else pari_err(impl,"more than 16 references");
    case OCpushptr:
        {
          gp_pointer *g = &ptrs[rp-1];
          gel(st,sp++) = (GEN)&(g->x); 
        }
        break;
    case OCendptr:
        for(j=0;j<operand;j++)
        {
          gp_pointer *g = &(ptrs[--rp]);
          if (g->ep) changevalue(g->ep, g->x);
          else (void)change_compo(avma, &(g->c), g->x);
        }
        break;
    case OCprecreal:
        st[sp++]=precreal;
        break;
    case OCprecdl:
        st[sp++]=precdl;
        break;
    case OCstoi:
        gel(st,sp-1)=stoi(st[sp-1]);
        break;
    case OCitos:
        st[sp-1]=itos(gel(st,sp-1));
        break;
    case OCtostr:
        if (operand==1)
          st[sp-1] = (long) GSTR(GENtoGENstr(gel(st,sp-1)));
        else
        {
          GEN L=cgetg(operand+1,t_VEC);
          sp-=operand;
          for (j=1; j<=operand; j++)
            gel(L,j) = GENtoGENstr(gel(st,sp-1+j));
          st[sp++] = (long) GSTR(concat(L,NULL));
        }
        break;
    case OCvarn:
        st[sp-1] = closure_varn(gel(st,sp-1));
        break;
    case OCpolx:
        gel(st,sp-1)=pol_x(st[sp-1]);
        break;
    case OCcompo1:
        {
          GEN  p=gel(st,sp-2);
          long c=st[sp-1];
          sp-=2;
          switch(typ(p))
          {
          case t_VEC: case t_COL:
            check_array_index(c, lg(p));
            closure_castgen(gel(p,c),operand);
            break;
          case t_LIST:
            check_array_index(c,lgeflist(p)-1);
            closure_castgen(gel(p,c+1),operand);
            break;
          case t_VECSMALL:
            check_array_index(c,lg(p));
            closure_castlong(p[c],operand);
            break;
          default:
            pari_err(talker,"[_]: not a vector");
            break;
          }
          break;
        }
    case OCcompo1ptr:
        { 
          long c=st[sp-1];
          gp_pointer *g = &ptrs[rp-1];
          matcomp *C=&g->c;
          GEN p;
          if (g->ep)
          {
            g->ep=NULL;
            C->full_col = C->full_row = 0;
            C->parent   = (GEN)    g->x;
            C->ptcell   = (GEN *) &g->x;
          }
          p=*C->ptcell;
          sp--;
          switch(typ(p))
          {
          case t_VEC: case t_COL: case t_VECSMALL:
            check_array_index(c, lg(p));
            C->ptcell = (GEN *) p+c;
            break;
          case t_LIST:
            check_array_index(c,lgeflist(p)-1);
            C->ptcell = (GEN *) p+c+1;
            break;
          default:
            pari_err(talker,"[_]: not a vector");
          }
          C->parent   = p;
          g->x = *(g->c.ptcell);
          break;
        }
    case OCcompo2:
        {
          GEN  p=gel(st,sp-3);
          long c=st[sp-2];
          long d=st[sp-1];
          if (typ(p)!=t_MAT)
            pari_err(talker,"[_,_]: not a matrix");
          check_array_index(d, lg(p));
          check_array_index(c, lg(p[d]));
          sp-=3;
          closure_castgen(gcoeff(p,c,d),operand);
          break;
        }
    case OCcompo2ptr:
        {
          long b=st[sp-2];
          long c=st[sp-1];
          gp_pointer *g = &ptrs[rp-1];
          matcomp *C=&g->c;
          GEN p;
          if (g->ep)
          {
            g->ep=NULL;
            C->full_col = C->full_row = 0;
            C->ptcell   = (GEN *) &g->x;
          }
          p=*C->ptcell;
          sp-=2;
          if (typ(p)!=t_MAT)
            pari_err(talker,"[_,_]: not a matrix");
          check_array_index(b, lg(p));
          check_array_index(c, lg(p[b]));
          C->ptcell = (GEN *) gel(p,c)+b;
          C->parent   = p;
          g->x = *(g->c.ptcell);
          break;
        }
    case OCcompoC:
        {
          GEN  p=gel(st,sp-2);
          long c=st[sp-1];
          if (typ(p)!=t_MAT)
            pari_err(talker,"[,_]: not a matrix");
          check_array_index(c, lg(p));
          sp--;
          gel(st,sp-1) = gcopy(gel(p,c));
          break;
        }
    case OCcompoCptr:
        {
          long c=st[sp-1];
          gp_pointer *g = &ptrs[rp-1];
          matcomp *C=&g->c;
          GEN p;
          if (g->ep)
          {
            g->ep=NULL;
            C->full_row = 0;
            C->ptcell   = (GEN *) &g->x;
          }
          p=*C->ptcell;
          sp--;
          if (typ(p)!=t_MAT)
            pari_err(talker,"[,_]: not a matrix");
          check_array_index(c, lg(p));
          C->ptcell = (GEN *) p+c;
          C->full_col = c;
          C->parent   = p;
          g->x = *(g->c.ptcell);
          break;
        }
    case OCcompoL:
        {
          GEN  p=gel(st,sp-2), p2;
          long c, r=st[sp-1];
          sp--;
          if (typ(p)!=t_MAT)
            pari_err(talker,"[_,]: not a matrix");
          if (lg(p)==1) pari_err(talker,"a 0x0 matrix has no elements");
          check_array_index(r,lg(p[1]));
          p2 = cgetg(lg(p),t_VEC);
          for (c=1; c<lg(p); c++) gel(p2,c) = gcoeff(p,r,c);
          gel(st,sp-1) = p2;
          break;
        }
    case OCcompoLptr:
        {
          long c,r=st[sp-1];
          gp_pointer *g = &ptrs[rp-1];
          matcomp *C=&g->c;
          GEN p, p2;
          if (g->ep)
          {
            g->ep=NULL;
            C->full_col = 0;
            C->ptcell   = (GEN *) &g->x;
          }
          p=*C->ptcell;
          sp--;
          if (typ(p)!=t_MAT)
            pari_err(talker,"[_,]: not a matrix");
          if (lg(p)==1) pari_err(talker,"a 0x0 matrix has no elements");
          check_array_index(r,lg(p[1]));
          p2 = cgetg(lg(p),t_VEC);
          C->full_row = r; /* record row number */
          for (c=1; c<lg(p); c++) gel(p2,c) = gcoeff(p,r,c);
          C->ptcell = &p2;
          C->parent   = p;
          g->x = *(g->c.ptcell);
          break;
        }
    case OCgetarg:
        ep=(entree *)operand;
        if (gel(st,sp-1))
          copyvalue(ep,gel(st,sp-1));
        else
          copyvalue(ep,gen_0);
        sp--;
        break;
    case OCdefaultarg:
        ep=(entree *)operand;
        if (gel(st,sp-2))
          copyvalue(ep,gel(st,sp-2));
        else
        {
          GEN z = closure_eval(gel(st,sp-1));
          if (!z) pari_err(talker,"break not allowed in args");
          copyvalue(ep,z);
        }
        sp-=2;
        break;
    case OCglobalvar:
        ep=(entree *)operand;
        ep->valence = EpGVAR;
        if (gel(st,sp-1))
          changevalue(ep,gel(st,sp-1));
        sp--;
        break;
#define ARGS st[sp],st[sp+1],st[sp+2],st[sp+3],st[sp+4],st[sp+5],st[sp+6],st[sp+7]
    case OCderivgen:
        {
          entree *ep=(entree*)operand;
          GEN res;
          struct derivgenwrap_s c;
          sp-=ep->arity;
          gp_function_name=ep->name;
          c.f = (GEN (*) (ANYARG)) ep->value;
          c.x = (GEN*) st+sp;
          res = derivnum((void*)&c, derivgenwrap, gel(st,sp), precreal);
          gp_function_name=NULL;
          gel(st,sp++)=res;
          break;
        }
    case OCcallgen:
        {
          entree *ep=(entree*)operand;
          GEN res;
          sp-=ep->arity;
          gp_function_name=ep->name;
          res = ((GEN (*)(ANYARG))ep->value)(ARGS);
          if (br_status) { avma=ltop; return NULL; }
          gp_function_name=NULL;
          gel(st,sp++)=res;
          break;
        }
    case OCcalllong:
        {
          entree *ep=(entree*)operand;
          long res;
          sp-=ep->arity;
          gp_function_name=ep->name;
          res = ((long (*)(ANYARG))ep->value)(ARGS);
          if (br_status) {avma=ltop; return NULL;}
          gp_function_name=NULL;
          st[sp++] = res;
          break;
        }
    case OCcallint:
        {
          entree *ep=(entree*)operand;
          long res;
          sp-=ep->arity;
          gp_function_name=ep->name;
          res = ((int (*)(ANYARG))ep->value)(ARGS);
          if (br_status) {avma=ltop; return NULL;}
          gp_function_name=NULL;
          st[sp++] = res;
          break;
        }
    case OCcallvoid:
        {
          entree *ep=(entree*)operand;
          sp-=ep->arity;
          gp_function_name=ep->name;
          ((void (*)(ANYARG))ep->value)(ARGS);
          if (br_status) {avma=ltop; return NULL;}
          gp_function_name=NULL;
          break;
        }
#undef ARGS
    case OCderivuser:
        {
          GEN z;
          long n=st[--sp];
          ep = (entree*) operand;
          if (ep->valence==EpNEW)
            pari_err(talker,"function '%s' not yet defined",ep->name);
          if (n>ep->arity)
            pari_err(talker,"Too many arguments for function '%s'",ep->name);
          for (j=n+1;j<=ep->arity;j++)
            gel(st,sp++)=0;
          z = derivnum((void*)ep, derivuserwrap, gel(st,sp-ep->arity), precreal);
          sp-=ep->arity;
          gel(st, sp++) = z;
          break;
        }
    case OCcalluser:
        {
          GEN z;
          long n=st[--sp];
          long savsp=sp;
          ep = (entree*) operand;
          if (ep->valence==EpNEW)
            pari_err(talker,"function '%s' not yet defined",ep->name);
          if (n>ep->arity)
            pari_err(talker,"Too many arguments for function '%s'",ep->name);
          for (j=n+1;j<=ep->arity;j++)
            gel(st,sp++)=0;
#ifdef STACK_CHECK
          if (PARI_stack_limit && (void*) &z <= PARI_stack_limit)
            pari_err(talker, "deep recursion");
#endif
          z = closure_eval((GEN) ep->value);
          if (br_status)
          {
            sp = savsp-n;
            br_status = br_NONE;
            if (br_res) z=gcopy(br_res);
            if (!z) z=gnil;
          }
          /* ep->value, beware: it may be killed anytime.  */
          if (isclone(z)) z = gcopy(z);
          for(j=1;j<lg(ep->lvars);j++)
            pop_val((entree*)ep->lvars[j]);
          gel(st, sp++) = z;
          break;
        }
    case OCvec:
    case OCcol:
    case OCmat:
        {
          long j;
          GEN v = cgetg(operand+1,
                       (opcode==OCvec?t_VEC:opcode==OCcol?t_COL:t_MAT));
          for (j=1;j<=operand;j++)
            gel(v,j)=copyifclone(gel(st,sp-operand+j-1));
          sp-=operand;
          gel(st,sp++)=v;
          break;
        }
    case OCdeffunc:
        ep=(entree*)operand;
        ep->value = (void *) gclone(gel(st,sp-4));
        ep->valence = EpUSER;
        ep->lvars = gclone(gel(st,sp-3));
        ep->code = pari_strdup(GSTR(gel(st,sp-2)));
        ep->arity = st[sp-1];
        sp-=4;
        gel(st, sp-1) = gnil;
        break;
    case OCpop:
        sp-=operand;
        break;
    }
  }
  sp--;
  if (sp<0) pari_err(bugparier,"closure_eval");
  if (gel(st,sp)==gnil) {avma=ltop; return gnil;}
  if (gel(st,sp)==NULL) pari_err(bugparier,"closure_eval");
  return gerepilecopy(ltop,gel(st,sp)); 
}

void
closure_reset(void) {sp=0;}

void
closure_check(void)
{
  if (sp) 
    pari_err(bugparier,"gp_closure, stack not empty");
}

INLINE const char *
disassemble_cast(long mode)
{
  switch (mode)
  {
  case Gsmall:
    return "small";
  case Ggen:
    return "gen";
  case Gvar:
    return "var";
  case Gvoid:
    return "void";
  default:
    return "unknown";
  }
}

void
closure_disassemble(GEN C)
{
  GEN code;
  long i;
  if (typ(C)!=t_VEC || lg(C)!=3 || typ(C[1])!=t_VECSMALL) 
    pari_err(typeer,"disassemble");
  code=gel(C,1);
  for(i=1;i<lg(code);i+=2)
  {
    op_code opcode=code[i];
    long operand=code[i+1];
    entree *ep;
    pariprintf("%05ld\t",i);
    switch(opcode)
    {
    case OCpushlong:
      if (operand==(long)gnil)
        pariprintf("pushlong\tgnil\n");
      else
        pariprintf("pushlong\t%ld\n",operand);
      break;
    case OCpushgen:
      pariprintf("pushgen\t\t%ld\n",operand);
      break;
    case OCpushreal:
      pariprintf("pushreal\t%ld\n",operand);
      break;
    case OCpushstoi:
      pariprintf("pushstoi\t%ld\n",operand);
      break;
    case OCpushentryval:
      ep=(entree*)operand;
      pariprintf("pushentryval\t%s\n",ep->name);
      break;
    case OCnewptr:
      ep=(entree*)operand;
      pariprintf("newptr\t\t%s\n",ep->name);
      break;
    case OCpushptr:
      pariprintf("pushptr\n");
      break;
    case OCendptr:
      pariprintf("endptr\t\t%ld\n",operand);
      break;
    case OCprecreal:
      pariprintf("precreal\n");
      break;
    case OCprecdl:
      pariprintf("precdl\n");
      break;
    case OCstoi:
      pariprintf("stoi\n");
      break;
    case OCitos:
      pariprintf("itos\n");
      break;
    case OCtostr:
      pariprintf("tostr\t\t%ld\n",operand);
      break;
    case OCvarn:
      pariprintf("varn\n");
      break;
    case OCpolx:
      pariprintf("polx\n");
      break;
    case OCcompo1:
      pariprintf("compo1\t\t%s\n",disassemble_cast(operand));
      break;
    case OCcompo1ptr:
      ep=(entree*)operand;
      pariprintf("compo1ptr\n");
      break;
    case OCcompo2:
      pariprintf("compo2\t\t%s\n",disassemble_cast(operand));
      break;
    case OCcompo2ptr:
      pariprintf("compo2ptr\n");
      break;
    case OCcompoC:
      pariprintf("compoC\n");
      break;
    case OCcompoCptr:
      pariprintf("compoCptr\n");
      break;
    case OCcompoL:
      pariprintf("compoL\n");
      break;
    case OCcompoLptr:
      pariprintf("compoLptr\n");
      break;
    case OCgetarg:
      ep=(entree*)operand;
      pariprintf("getarg\t\t%s\n",ep->name);
      break;
    case OCdefaultarg:
      ep=(entree*)operand;
      pariprintf("defaultarg\t%s\n",ep->name);
      break;
    case OCglobalvar:
      ep=(entree*)operand;
      pariprintf("globalvar\t%s\n",ep->name);
      break;
    case OCderivgen:
      ep=(entree*)operand;
      pariprintf("derivgen\t\t%s\n",ep->name);
      break;
    case OCcallgen:
      ep=(entree*)operand;
      pariprintf("callgen\t\t%s\n",ep->name);
      break;
    case OCcalllong:
      ep=(entree*)operand;
      pariprintf("calllong\t%s\n",ep->name);
      break;
    case OCcallint:
      ep=(entree*)operand;
      pariprintf("callint\t\t%s\n",ep->name);
      break;
    case OCcallvoid:
      ep=(entree*)operand;
      pariprintf("callvoid\t%s\n",ep->name);
      break;
    case OCderivuser:
      ep=(entree*)operand;
      pariprintf("derivuser\t\t%s\n",ep->name);
      break;
    case OCcalluser:
      ep=(entree*)operand;
      pariprintf("calluser\t%s\n",ep->name);
      break;
    case OCvec:
    case OCcol:
    case OCmat:
      pariprintf("%s\t\t%ld\n",
          (opcode==OCvec?"vec":opcode==OCcol?"col":"mat"),operand);
      break;
    case OCdeffunc:
      ep=(entree*)operand;
      pariprintf("deffunc\t\t%s\n",ep->name);
      break;
    case OCpop:
      pariprintf("pop\t\t%ld\n",operand);
      break;
    }
  }
}
